// Pawe Zubrycki, K3I3
//	Algorytm for_each

/*	Algorytm for_each umoliwia dostp, przetworzenie i modyfikacj kadego
	elementu kontenera, w tym rwnie zwykej tablicy. Zaliczany jest do 
	algorytmw niemodyfikujcych, ale moe by te zaliczany do algorytmw 
	modyfikujcych.
	Algorytm for_each jest uywany zamiast ptli for, aby uzwili i poprawi
	czytelno kodu przy iteracji przez kontenery. Jego uycie jednoznacznie 
	okrela intencje autora i ogranicza rozwleko kodu. 

	Nagwek:			*/
#		include<algorithm>
/*			
	Deklaracja for_each:
		template<class InputIterator, class UnaryProc>
		UnaryProc for_each(InputIterator begin, InputIterator end, UnaryProc operation)

	Parametry:
		begin		- pierwszy element kontenera
		end		- element za ostatnim w kontenerze
		operation	- operacja wykonywana na elementach kontenera. Moe by obiektem funkcyjnym
	
	Warto zwracana:
		Kopia operation. Jeli operation byo obiektem funkcyjnym zmieniajcym stan 
		podczas iteracji, for_each zwrci ten obiekt po przetworzeniu.
	
	Dziaanie:
		Wywouje operation(elem) dla kadego elementu z zakresu [begin, end). operation 
		moe by zwykym wskanikiem na funkcj, jak i obiektem funkcyjnym. Algorytm po
		przetworzeniu zwraca operation.

	Nagwki pomocnicze:		
*/
#include<iostream>
#include<string>
#include<map>
#include<boost/lambda/lambda.hpp>
#include<boost/lambda/bind.hpp>

using namespace std;
using namespace boost::lambda;
namespace b = boost;
		
/*/ Algorytm for_each jako algorytm niemodyfikujcy: 
	
	for_each moe zosta uyte do wypisania wszystkich elementw kontenera.
	
	W przykadzie posuymy si algorytmem for_each do wyliczenia redniej waonej. Skorzystamy z tego, 
	e for_each zwraca operacj, ktra moe by przetworzonym obiektem funkcyjnym.
*/
//	Obiekt funkcyjny obliczajcy redni waon elementw multimapy zapisanych w formacie [waga, liczba]
class Weighted_mean{
	double value_;
	unsigned int n_;
public:
	Weighted_mean(): value_(0.0), n_(0){}
	void operator()(pair<unsigned int, double> factor){ 
		value_ += factor.first * factor.second;
	    	n_ += factor.first;
	}
	operator double(){
		if(n_ == 0) return 0.0;
		return value_/n_;
	}
};

/*/ Algorytm for_each jako algorytm modyfikujcy:

	Kolejny przykad pokazuje jak mona zmodyfikowa zawarto kontenera, np. chcc doda podatek
	do produktw:
*/
//	Obiekt funkcyjny sucy do wypisania produktw 
class PrintProducts{
	unsigned int count_;	
public:
	// Konstruktor przyjmuje parametr wywietlany przed list produktw
	explicit PrintProducts(std::string s = ""): count_(0) { if(!s.empty()) cout << s << endl; }
	void operator()(pair<string, double> a){
		cout << ++count_ << ". " << a.first << "\tcena: " << a.second << "zl" << endl; 
	}
};
	
int main(){
	locale pl("pl_PL.ISO-8859-2");
	cout.imbue(pl);
	
	cout << "PRZYKAD 1\n" << endl;
	// Multimapa ocen, gdzie kluczami s wagi
	typedef multimap<unsigned int, double> Oceny;
	Oceny oceny;
	oceny.insert(Oceny::value_type(4, 3.0));// pary [waga, ocena]
	oceny.insert(Oceny::value_type(3, 5.0));
	oceny.insert(Oceny::value_type(1, 5.0));
	cout << "rednia ocen: \n";
	// Aby nie tworzy prostych obiektw funkcyjnych jednorazowego uytku mona uy "magii" biblioteki boost:lambda 
	for_each(oceny.begin(), oceny.end(), 
		cout << constant("Waga: ") << bind(&Oceny::value_type::first, _1) << constant(", Ocena: ") 
			<< bind(&Oceny::value_type::second, _1) << constant('\n')); 

	cout << "\nwynosi: " << static_cast<double>(for_each(oceny.begin(), oceny.end(), Weighted_mean())) << endl;
	cout << "\nPRZYKAD 2\n" << endl;	
	
	typedef map<string, double> map_type;
	map_type koszyk;

	//Dodajemy produkty
	koszyk.insert(map_type::value_type("skarpety", 5.0));
	koszyk.insert(map_type::value_type("jeansy", 149.99));
	//Ustawiamy parametry strumienia wyjciowego
	cout.precision(2);
	cout.setf(ios::fixed);
	for_each(koszyk.begin(), koszyk.end(), PrintProducts("Ceny netto: "));
	// W tym miejscu znowu korzystamy z boost::lambda. Mnoymy kad cen(skadowa pair<string, double>::second) przez 1.22  
	for_each(koszyk.begin(), koszyk.end(), (bind( &map_type::value_type::second, _1) *= 1.22));
	for_each(koszyk.begin(), koszyk.end(), PrintProducts("Ceny brutto: "));
	return 0;
}

